/*
 * fourier.c
 *
 * Peter Mawhorter 2009-1-6
 *
 * D
 */

#include <fftw3.h>
#include <stdlib.h>
#include <string.h>

#include "fourier.h"

/*********************
 * Global variables: *
 *********************/

fftw_complex *in, *out, *buffer; // Arrays for the FFT data
fftw_plan p; // The plan for the FFT

int leftovers = 0; // The end of old data in the buffer.

float frameduration = 0.02; // The duration of a frame in seconds (must be
                            // smaller than the duration of a period).

int framesamples = 0; // The number of samples in a frame.
int periodsamples = 0; // The number of samples in a period.

// Variables copied from arguments to setupFourier:
int psize;
int chnls;
int bps;
int end;
int rte;

// The (normalized) power necessary to qualify as non-silence:
float silence_threshold = 0.01;

/**************
 * Functions: *
 **************/

/*
 * This function sets up for the Fourier transforms to come, creating a plan
 * for the transform and allocating arrays for the data. It also sets up the
 * global variables that regulate the data transform.
 */
void setupFourier(int periodsize, int channels, int bytespersample,
                  int endianness, int rate)
{
  framesamples = frameduration * rate;
  periodsamples = periodsize / (channels * bytespersample);
  psize = periodsize;
  chnls = channels;
  bps = bytespersample;
  end = endianness;
  rte = rate;
  buffer = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * periodsamples*2);
  in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * framesamples);
  out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * framesamples);
  p = fftw_plan_dft_1d(framesamples, in, out, FFTW_FORWARD, FFTW_MEASURE);
}

/*
 * Destroys the plan and frees the arrays that were created in setupFourier.
 */
void teardownFourier()
{
  fftw_destroy_plan(p);
  fftw_free(buffer);
  fftw_free(in);
  fftw_free(out);
}

/*
 * This function moves data from a uchar array (where it is packed according
 * to the known period size, number of channels, bytes per sample, and
 * endianness) into a fftw_complex array in preparation for the FFT.
 */
void transferData(unsigned char* data, fftw_complex* destination, int bindex)
{
  int i, j;
  int index;
  int value;
  if (end) { // Little-endian
    for (i = 0; i < periodsamples; ++i) {
      index = i * bps * chnls;
      value = 0;
      for (j = 0; j < bps; ++j) {
        value += data[index+j] << (8 * j);
      }
      j--;
      value -= (data[index + j] & 128) << (8 * j + 1); // Signed!
      // DEBUG:
      // printf("%d\n", value);
      destination[bindex + i][0] = value;
      destination[bindex + i][1] = 0;
    }
  } else { // Big-endian
    for (i = 0; i < periodsamples; ++i) {
      index = i * bps * chnls;
      value = 0;
      value -= (data[index] & 128) << (8 * (bps - 1) + 1); // Signed!
      for (j = 0; j < bps; ++j) {
        value += data[index+j] << (8 * (bps - j - 1));
      }
      // DEBUG:
      // printf("%d\n", value);
      destination[bindex + i][0] = value;
      destination[bindex + i][1] = 0;
    }
  }
}

/*
 * This method takes a signal and returns true if it meets the given threshold
 * and false otherwise. The threshold should be on [0, 1) and specifies a
 * normalized power constraint.
 */
int meetsThreshold(fftw_complex *signal, int bps, float threshold)
{
  int i;
  int power = 0;
  for (i = 0; i < framesamples; ++i) {
    if (abs(signal[i][0]) > power) {
      power = abs(signal[i][0]);
    }
  }
  // DEBUG:
  //printf("%.3f\n", ((float) power / (float) (1 << (bps * 8 - 1))));
  return ((float) power / (float) (1 << (bps * 8 - 1))) > threshold;
}

/*
 * This function processes a period of sound data. It computes the FFT of each
 * frame of data and finds the peak frequency of each frame (using 0 if the
 * frame doesn't pass the silence threshold). It returns these frequencies as
 * a mallocd array of integers, the first of which contains the length. This
 * array should be freed by the code which processes the result.
 */
int* processPeriod(unsigned char *data)
{
  int i, lastodd, lasteven, framesthisperiod;
  int *peaks;
  // Calculate number of frames (includes even and odd frames):
  framesthisperiod = ((periodsamples + leftovers) / framesamples) + \
                     ((periodsamples + leftovers - (framesamples / 2)) /\
                      framesamples);

  // Malloc for the peaks:
  peaks = (int*) malloc(sizeof(int) * (1 + framesthisperiod));

  // Set the number of peaks:
  peaks[0] = framesthisperiod;

  // DEBUG:
  // printf("P%d \n", peaks[0]);

  // DEBUG:
  // for (i = 0; i < periodsamples; i += 2) {
  //   printf("%d: %d\n", i, data[i] + (data[i + 1] << 8) -
  //                         ((data[i + 1] & 128) << 9));
  // }

  // Move that data into the buffer after any leftovers (and convert it):
  transferData(data, buffer, leftovers);

  // DEBUG:
  // for (i = 0; i < periodsamples + leftovers; ++i) {
  //   printf("%d: %.2f\n", i, buffer[i][0]);
  // }

  lasteven = 0;
  lastodd = 0;
  // Process each frame of the period:
  for(i = 0; i < framesthisperiod; ++i) {
    // Grab the frame data:
    if (i % 2) {
      memcpy(in, &(buffer[(i/2) * framesamples + framesamples / 2]),
             sizeof(fftw_complex) * framesamples);
      lastodd = (i/2) * framesamples + (3 * framesamples / 2);
    } else {
      memcpy(in, &(buffer[(i/2) * framesamples]),
             sizeof(fftw_complex) * framesamples);
      lasteven = (i/2) * framesamples + framesamples;
    }

    // DEBUG:
    // for (int j = 0; j < framesamples; ++j) {
    //   printf("%d: %.2f\n", j, in[j][0]);
    // }

    // FFT it:
    fftw_execute(p);

    // Threshold and compute peak:
    if (meetsThreshold(in, bps, silence_threshold)) {
      peaks[i + 1] = findPeakFrequency(out, frameduration);
      // DEBUG:
      // printf("#%d: %d\n", i, peaks[i]);
    } else {
      peaks[i + 1] = 0;
      // DEBUG:
      // printf("#%d, %d\n", i, peaks[i]);
    }
  }

  // Move leftovers to the beginning of the buffer:
  if (lasteven < lastodd) {
    memmove(buffer, &(buffer[lasteven]),
            sizeof(fftw_complex) * (periodsamples + leftovers - lasteven));
    leftovers = periodsamples + leftovers - lasteven;
  } else {
    memmove(buffer, &(buffer[lastodd]),
            sizeof(fftw_complex) * (periodsamples + leftovers - lastodd));
    leftovers = periodsamples + leftovers - lastodd;
  }
  // DEBUG:
  // printf("\n");

  return peaks;
}

/*
 * This method finds the peak frequency of a signal, given its Fourier
 * transform and duration.
 */
int findPeakFrequency(fftw_complex *transform, float duration)
{
  int k;
  int middle;
  double frequency;
  double power;
  double peak;
  double peakpower;
  // Figure out what the highest-frequency power is (it might be divided) and
  // compute the middle index (use the computed power as the beginning of our
  // max calculation):
  if (framesamples % 2) {
    middle = (framesamples - 1) / 2;
    peakpower = abs(transform[middle][0]);
    peak = middle / duration;
  } else {
    middle = framesamples / 2;
    peakpower = abs(transform[middle][0]) / 2.0;
    peak = middle / duration;
  }
  // Process all of the frequencies between the DC power level (index 0) and
  // the highest recognizable frequency (k/n) which is computed above:
  for (k = 1; k < middle; ++k) {
    frequency = k / duration;
    power = abs(transform[k][0]);
    if (power > peakpower) {
      peakpower = power;
      peak = frequency;
    }
  }
  return (int) peak;
}
