'''
recognize.py
Peter Mawhorter
2008-1-8
This module implements a note-sequence recognition algorithm given a list of
peak frequencies. It gets this list real-time from a queue that is fed by the
listen method in navi.py.
'''

import math, os

sameNoteThreshold = 1.0 # Half-steps of difference tolerance.
stepsTolerance = 1.0 # Half-steps of recognition tolerance.

# superframesize = 12 # The number of frames per superframe (must be even).
superframesize = 6 # The number of frames per superframe (must be even).
buffer = [0]*superframesize # The buffer for storing superframes.
superframeindex = 0 # The index into the buffer.

# maxpause = 15 # Superframes of pause allowed before rest.
maxpause = 40 # Superframes of pause allowed before rest.
superslop = 2 # Number of slop superframes allowed.
slopused = 0 # Superslop used so far.
slop = 1 # Number of slop frames allowed.

last = 0 # The pitch of the last note.

# A dictionary for translating from letters to relative pitches:
lnums = {
'a':0,
'b':2,
'c':3,
'd':5,
'e':7,
'f':8,
'g':10
}

# A dictionary for translating from pitches to letters:
nletters = {
0:'a',
1:'a#',
2:'b',
3:'c',
4:'c#',
5:'d',
6:'d#',
7:'e',
8:'f',
9:'f#',
10:'g',
11:'g#'
}

def attend(queue):
  '''
  Activates recognition of frequencies inserted into the given queue.
  '''
  global buffer, superframeindex
  pitch = 0 # Tracks last superframe pitch.
  duration = 0 # Counts note superframes.
  while True:
    item = queue.get()
    if item == 'Q': # The quit signal
      break
    elif type(item) == int: # A frequency
      # DEBUG:
      #print item
      buffer[superframeindex] = item
      superframeindex += 1
      if superframeindex >= len(buffer): # A full superframe is in: process it.
        # DEBUG:
        #print buffer
        pitch, duration = processSuperframe(buffer, pitch, duration)
        # Finally, chop off the first half of the superframe and get ready for
        # the next (overlapping) superframe.
        buffer = buffer[len(buffer)/2:] + [0]*(len(buffer)/2)
        superframeindex = len(buffer)/2

def processSuperframe(buffer, pitch, duration):
  '''
  Analyses a superframe in an attempt to classify it as either silence, a note
  at a certain pitch, or a transition between notes. Returns new duration and
  pitch values, and may call gotnote.
  '''
  # Cluster the pitches (recording indices for each cluster):
  clusters = {}
  for i in range(len(buffer)):
    if buffer[i] not in clusters: # A new note:
      hit = False
      # Does this note belong in an existing cluster?:
      for c in clusters:
        if abs(sd(c, buffer[i])) <= sameNoteThreshold:
          hit = True
          # Weighted average update:
          clusters[(c * len(clusters[c]) + buffer[i]) /
                   float(len(clusters[c]) + 1)] = clusters[c] + [i]
          # Get rid of the old (pre-averaging) cluster
          del clusters[c]
          break
      # No.. it doesn't. make a new cluster for it:
      if not hit:
        clusters[buffer[i]] = [i]
    else: # An old note: add it to it's cluster:
      clusters[buffer[i]] = clusters[buffer[i]] + [i]

  # Remove slop clusters:
  condemned = []
  for c in clusters:
    if len(clusters[c]) <= slop:
      # Test for harmonic:
      for f in clusters:
        if abs(sd(c, f*2)) <= sameNoteThreshold*2 or\
           abs(sd(c, f*4)) <= sameNoteThreshold*4:
          # Credit this slop to the cluster it's a harmonic of:
          clusters[f] = clusters[f] + clusters[c]
          break
      condemned.append(c)
  for c in condemned:
    if len(clusters[c]) <= slop: # Test again for slop in light of harmonics:
      del clusters[c]

  # See if it's an easy decision:
  if len(clusters) == 1 and len(clusters[clusters.keys()[0]]) >= len(buffer)/2:
    return processPitch(clusters.keys()[0], pitch, duration)
  elif len(clusters) == 1:
    return processPitch("noise", pitch, duration)

  # Aww shucks. Test for harmonic duality:
  if len(clusters) == 2 and\
     (abs(sd(clusters.keys()[0], 2 * clusters.keys()[1])) <= sameNoteThreshold\
   or abs(sd(clusters.keys()[1], 2 * clusters.keys()[0])) <= sameNoteThreshold)\
  and abs(sum(clusters.values()[0])/float(len(clusters.values()[0])) - \
          sum(clusters.values()[1])/float(len(clusters.values()[1]))) <=\
      len(buffer)/2:
    return processPitch(min(clusters.keys()), pitch, duration)

  # Well... is there a dominant cluster?
  maxlen = 0
  subsubmaxlen = 0
  submaxlen = 0
  subsubdominant = 0
  subdominant = 0
  dominant = 0
  for c in clusters:
    if len(clusters[c]) > maxlen:
      subsubmaxlen = submaxlen
      subsubdominant = subdominant
      submaxlen = maxlen
      subdominant = dominant
      maxlen = len(clusters[c])
      dominant = c

  if maxlen > submaxlen*2 and maxlen >= len(buffer)/2:
    return processPitch(dominant, pitch, duration) # A single pitch is dominant

  if maxlen < submaxlen * 2 and submaxlen > subsubmaxlen * 2:
    # Two pitches dominate together
    maxposition = sum(clusters[dominant]) /\
                  float(len(clusters[dominant]))
    submaxposition = sum(clusters[subdominant]) /\
                     float(len(clusters[subdominant]))
    # Does this look like a transition?
    if maxposition < len(buffer)/4 + 1 and\
       submaxposition > 3 * len(buffer)/4 - 1:
      return processPitch((dominant, subdominant), pitch, duration)
    elif submaxposition < len(buffer)/4 + 1 and\
         maxposition > 3 * len(buffer)/4 - 1:
      return processPitch((subdominant, dominant), pitch, duration)
    # Are the pitches harmonics of each other?
    if (abs(sd(dominant, 2 * subdominant)) <= 2 * sameNoteThreshold or \
        abs(sd(subdominant, 2 * dominant)) <= 2 * sameNoteThreshold) and\
       abs(maxposition - submaxposition) <= len(buffer)/2:
      return processPitch(min(dominant, subdominant), pitch, duration)

  # Three or more pitches... ugh.

  # Maybe it's an ascent or descent:
  # TODO: Throw out outliers in the rate average!
  ascent = 0
  arate = 0
  apos = 0
  descent = 0
  drate = 0
  dpos = 0
  for i in range(1, len(buffer)):
    steps = sd(buffer[i], buffer[i - 1])
    if abs(steps) > sameNoteThreshold:
      if steps > 0:
        ascent += 1
        # Compute the running weighted averages for rate and position:
        arate += (steps - arate)/float(ascent)
        apos += (i - apos)/float(ascent)
      else:
        descent += 1
        # Compute the running weighted averages for rate and position:
        drate += (steps - drate)/float(descent)
        dpos += (i - dpos)/float(descent)

  if ascent >= 3 * len(buffer)/4 and abs(apos - dpos) <= len(buffer)/3:
    return processPitch(("ascent", min(clusters.keys()), arate * len(buffer)),
                        pitch, duration)
  elif descent >= 3 * len(buffer)/4 and abs(apos - dpos) <= len(buffer)/3:
    return processPitch(("descent", max(clusters.keys()),
                         drate * len(buffer)),
                        pitch, duration)

  return processPitch("hmmm...", pitch, duration)

def processPitch(incomming, old, duration):
  '''
  This function does the larger-scale recognition work of processSuperframe,
  taking the pitch for the current superframe (in any of the whacky ways that
  processSuperframe might pass it on) and trying to figure out what that means
  for the current note, of the given old (pitch) and duration. This function
  makes calls to gotnote as appropriate, and returns the new pitch and duration
  to hand back to the caller of processSuperframe.
  '''
  global slopused
  # DEBUG:
  #print '---', incomming, '-- %d ---' % int(old)
  newpitch = None
  newduration = None
  if incomming == "noise" or incomming == "hmmm...":
    # processSuperframe doesn't have a lock on the pitch. We'll ignore
    # superslop of these but then break the note.
    slopused += 1
    duration += 1
  elif type(incomming) == tuple:
    # Something special
    if incomming[0] == "ascent" or incomming[0] == "descent":
      # A gradual ascent or descent
      start = incomming[1]
      span = incomming[2]
      if abs(sd(start, old)) <= sameNoteThreshold * 2 or\
         abs(sd(start, old * (2**(span/12.0)))) <= sameNoteThreshold * 2:
        # An ascent or descent that starts on (or from) our note!
        gotnote(old, duration)
        newpitch = start * (2**(span/12.0))
        newduration = 0
      else:
        # Throw it into the slop bucket:
        slopused += 1
        duration += 1
    elif type(incomming[0]) == int and type(incomming[1]) == int:
      # A note transition
      if abs(sd(incomming[0], old)) <= sameNoteThreshold * 2:
        # A transition that starts on our note!
        gotnote(old, duration)
        newpitch = incomming[1]
        newduration = 1
      else:
        # Throw it into the slop bucket:
        slopused += 1
        duration += 1
  elif type(incomming) == int or type(incomming) == float:
    # Just a simple note
    if abs(sd(incomming, old)) <= sameNoteThreshold:
      # Average the note:
      slopused = 0
      newduration = duration + 1
      newpitch = old + (incomming - old)/float(newduration)
      # Handle the pause limit (return a rest every maxpause superframes):
      if incomming == 0 and newduration > maxpause:
        gotnote(newpitch, newduration)
        newduration = 0
    else:
      # A new note (superslop doesn't apply to well-recognized notes)
      gotnote(old, duration)
      slopused = 0
      newduration = 1
      newpitch = incomming
  else:
    # An unrecognized something!
    slopused += 1
    duration += 1

  # Handle slop overrun:
  if slopused > superslop:
    slopused = 0
    gotnote(old, duration)

  # Update duration and pitch if necessary:
  if newpitch != None:
    old = newpitch
  if newduration != None:
    duration = newduration
  return old, duration

def sd(pitch1, pitch2):
  '''
  This function computes the steps of difference between two pitches. It
  pretends that silence is 1 Hz. Just as if it were pitch1 - pitch2, the
  difference is positive if pitch1 is greater than pitch2 and negative
  (or zero) otherwise.
  '''
  if pitch1 == 0:
    pitch1 = 1
  if pitch2 == 0:
    pitch2 = 1
  return math.log(pitch1/float(pitch2), 2) * 12

def gotnote(pitch, duration):
  '''
  Called whenever a note is definitively recognized in the stream, this
  function is responsible for feeding the note to the phrases in hopes of
  recognition.
  '''
  global last

  # Filter out notes that are too short:
  if duration <= superslop:
    return
  if pitch == 0 and duration < maxpause:
    return

  # DEBUG:
  print ' ',
  printletter(pitch)

  if pitch > 0:
    if last > 0:
      # Calculate half-steps difference:
      steps = sd(pitch, last)

      # DEBUG:
      print "%.2f" % steps,

      if abs(steps) > sameNoteThreshold:
        for p in phrases:
          if p.feed(steps):
            if p.name == "Zelda's Lullaby":
              os.system('xdotool key "shift+F9"')
            print p.name
    else:
      for p in phrases:
        if p.feed("<"): # Exiting a rest
          print p.name
    last = pitch
  else:
    for p in phrases:
      if p.feed(">"): # Entering a rest
        print p.name
    last = 0

  # DEBUG:
  print

def printletter(pitch):
  '''
  This function takes a pitch and prints the corresponding letter.
  '''
  if pitch > 0:
    absolute = int(round(sd(pitch, 440)))
    while absolute >= 12:
      absolute -= 12
    while absolute < 0:
      absolute += 12
    print nletters[absolute],
  else:
    print "-",

class Phrase(object):
  '''
  A Phrase is a series of musical notes that can be recognized. It consists of
  a sequence of half-step difference values between successive notes, so that
  it can recognize the sequence independent of register or key. A Phrase object
  is also a state machine that can be fed note differences using the feed()
  function, which returns True when the phrase has been satisfied by a sequence
  of notes.
  '''
  def __init__(self, **kwargs):
    '''
    The constructor takes the following keyword arguments:
      name - a string identifying the phrase in some way.
      notes - a string of note letters with the following syntax:
                1st character: a letter of the scale (a-g) or one of "-" or
                               "?". "-" denotes a rest, "?" denotes an optional
                               rest. Multiple optional rests in a row are not
                               allowed, and the last note may not be an
                               optional rest.
                Additional characters: any of "#b',". "#" and "b" denote sharps
                                       and flats respectively. For each "'",
                                       the note is an octave higher relative to
                                       the current octave. For each ",", it's
                                       an octave lower. The current octave is
                                       the octave of the last note.


              An example (the tune here is Zelda's Lullaby):
                "b d a ? b d a"
              Another example (this time, the Song of Storms):
                "d f d' d, f d'"
              Note that after "'", "," is necessary to return to the old
              octave.
              A final example from Zelda's Lullaby:
                "d c b c b g, b' a g, g f# f#"
      steps - a list of step differences. Only one of notes or steps should be
              given. If both are given, notes is ignored in favor of steps. A
              ">" in this list signifies a transition to a rest, while a "<"
              signifies a transition from a rest. Use repeated ">"s for a
              serial rest. Use "?" for an optional transition to a rest, and
              make the next transition as if there were no rest. Note that
              relative frequencies of notes next to optional rests are not
              enforced. Multiple optional rests in a row are not allowed, and
              the last transition may not be an optional rest.
      tolerance - The tolerance, in half-steps for errors. Default is 1.0.
    '''
    defaults = {"tolerance":1.0, "steps":None, "notes":None}
    defaults.update(kwargs)
    self.name = defaults["name"]
    self.notes = defaults["notes"]
    self.steps = defaults["steps"]
    self.tolerance = defaults["tolerance"]
    if self.notes and not self.steps:
      letters = self.notes.split()
      numbers = []
      octave = 0
      for l in letters:
        accident = 0
        for i in range(1, len(l)):
          if l[i] == "'":
            octave += 1
          elif l[i] == ',':
            octave -= 1
          elif l[i] == '#':
            accident = 1
          elif l[i] == 'b':
            accident = -1
        if l[0] == '-':
          numbers.append('-')
        elif l[0] == '?':
          numbers.append('?')
        else:
          numbers.append(lnums[l[0]] + accident + 12*octave)
      self.steps = []
      for i in range(1, len(numbers)):
        if type(numbers[i]) == int:
          if type(numbers[i - 1]) == int:
            self.steps.append(numbers[i] - numbers[i - 1])
          else:
            if numbers[i - 1] == "-":
              self.steps.append("<")
            elif numbers[i - 1] == "?":
              if type(numbers[i - 2]) == int:
                self.steps.append(numbers[i] - numbers[i - 2])
              else:
                self.steps.append("<")
        else:
          if numbers[i] == "-":
            self.steps.append(">")
          elif numbers[i] == "?":
            self.steps.append("?")
    if not self.notes:
      self.notes = "some notes"
    self.atstates = [True] + [False]*(len(self.steps))

  def __str__(self):
    return self.name + ": " + self.notes

  def __repr__(self):
    return "Phrase(name=" + repr(self.name) + ", "\
                  "notes=" + repr(self.notes) + ", "\
                  "steps=" + repr(self.steps) + ", "\
                  "tolerance=" + repr(self.tolerance) + ")"

  def feed(self, steps):
    '''
    This method implements a state machine. It returns False unless the steps
    fed was the last steps needed to recognize the phrase, in which case it
    returns True.
    '''
    for i in range(len(self.steps) - 1, -1, -1):
      if self.atstates[i]:
        self.atstates[i] = False
        target = self.steps[i]
        if type(target) == str:
          if target == "?":
            if steps == ">":
              self.atstates[i + 1] = True
            elif type(steps) == float or type(steps) == int and\
                 abs(self.steps[i + 1] - steps) < self.tolerance:
              self.atstates[i + 2] = True
          elif target == "-":
            if target == steps:
                self.atstates[i+1] = True
        else:
          if (type(steps) == float or type(steps) == int) and\
             abs(target - steps) < self.tolerance:
            self.atstates[i+1] = True
          if i > 0 and self.steps[i - 1] == "?" and steps == "<":
            self.atstates[i+1] = True
    self.atstates[0] = True
    if self.atstates[-1]:
      self.reset()
      return True

  def reset(self):
    '''
    This function resets the state machine, and gets called whenever
    recognition succeeds or fails.
    '''
    self.atstates = [True] + [False]*(len(self.steps))

phrases = [
  Phrase(name="Zelda's Lullaby", notes="b d a ? b d a", tolerance=1.0),
  Phrase(name="Song of Storms", notes="d f d' ? d, f d'", tolerance=2.0),
  Phrase(name="Saria's Song", notes="f a' b ? f, a' b", tolerance=1.5)
]
